@@ -48,7 +48,8 @@ mperr(const char *format, ...)
 }
 
 
-const char *mp_get_error()
+const char *
+mp_get_error()
 {
     return mp_error;
 }
@@ -106,13 +107,13 @@ mp_get_eulers(MPNumber *z)
 }
 
 
-void mp_get_i(MPNumber *z)
+void
+mp_get_i(MPNumber *z)
 {
     mp_set_from_integer(0, z);
-    //memset(z, 0, sizeof(MPNumber));
-    //z->im_sign = 1;
-    //z->im_exponent = 1;
-    //z->im_fraction[0] = 1;
+    z->im_sign = 1;
+    z->im_exponent = 1;
+    z->im_fraction[0] = 1;
 }
 
 
@@ -121,8 +122,10 @@ mp_abs(const MPNumber *x, MPNumber *z)
 {
     if (mp_is_complex(x)){
         MPNumber x_real, x_im;
+
         mp_real_component(x, &x_real);
         mp_imaginary_component(x, &x_im);
+
         mp_multiply(&x_real, &x_real, &x_real);
         mp_multiply(&x_im, &x_im, &x_im);
         mp_add(&x_real, &x_im, z);
@@ -139,28 +142,46 @@ mp_abs(const MPNumber *x, MPNumber *z)
 void
 mp_arg(const MPNumber *x, MPAngleUnit unit, MPNumber *z)
 {
-    MPNumber x_real, x_im, t;
-    
+    MPNumber x_real, x_im, pi;
+
+    if (mp_is_zero(x)) {
+        /* Translators: Error display when attempting to take argument of zero */
+        mperr(_("Argument not defined for zero"));
+        mp_set_from_integer(0, z);
+        return;
+    }
+
     mp_real_component(x, &x_real);
     mp_imaginary_component(x, &x_im);
+    mp_get_pi(&pi);
 
-    if (mp_is_zero(&x_real)) {
-        mp_get_pi(z);
-        mp_divide_integer(z, 2, z);
-        if (mp_is_negative(&x_im)){
-            mp_get_pi(&t);
-            mp_add(z, &t, z);
-        }
+    if (mp_is_zero(&x_im)) {
+        if (mp_is_negative(&x_real))
+            convert_from_radians(&pi, MP_RADIANS, z);
+        else
+            mp_set_from_integer(0, z);
+    }
+    else if (mp_is_zero(&x_real)) {
+        mp_set_from_mp(&pi, z);
+        if (mp_is_negative(&x_im))
+            mp_divide_integer(z, -2, z);
+        else
+            mp_divide_integer(z, 2, z);
+    }
+    else if (mp_is_negative(&x_real)) {
+        mp_divide(&x_im, &x_real, z);
+        mp_atan(z, MP_RADIANS, z);
+        if (mp_is_negative(&x_im))
+            mp_subtract(z, &pi, z);
+        else
+            mp_add(z, &pi, z);
     }
     else {
-        mp_divide(&x_im, &x_real, &t);
-        mp_atan(&t, MP_RADIANS, z);
-
-        if (mp_is_negative(&x_real)) {
-            mp_get_pi(&t);
-            mp_add(z, &t, z);
-        }
+        mp_divide(&x_im, &x_real, z);
+        mp_atan(z, MP_RADIANS, z);
     }
+
+    convert_from_radians(z, unit, z);
 }
 
 
@@ -168,7 +189,7 @@ void
 mp_conjugate(const MPNumber *x, MPNumber *z)
 {
     mp_set_from_mp(x, z);
-    //z->im_sign = -z->im_sign;
+    z->im_sign = -z->im_sign;
 }
 
 
@@ -176,102 +197,105 @@ void
 mp_real_component(const MPNumber *x, MPNumber *z)
 {
     mp_set_from_mp(x, z);
-    //z->im_sign = 0;
-    //z->im_exponent = 0;
-    //memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);
+
+    /* Clear imaginary component */
+    z->im_sign = 0;
+    z->im_exponent = 0;
+    memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);
 }
 
 
 void
 mp_imaginary_component(const MPNumber *x, MPNumber *z)
 {
-    mp_set_from_integer(0, z);
-    //mp_set_from_mp(x, z);
-    //z->sign = z->im_sign;
-    //z->exponent = z->im_exponent;
-    //memcpy(z->fraction, z->im_fraction, sizeof(int) * MP_SIZE);
-    //z->im_sign = 0;
-    //z->im_exponent = 0;
-    //memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);
+    /* Copy imaginary component to real component */
+    z->sign = x->im_sign;
+    z->exponent = x->im_exponent;
+    memcpy(z->fraction, x->im_fraction, sizeof(int) * MP_SIZE);
+  
+    /* Clear (old) imaginary component */   
+    z->im_sign = 0;
+    z->im_exponent = 0;
+    memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);
 }
 
 
 static void
-mp_add_real(int x_sign, int x_exponent, const int *x_fraction,
-            int y_sign, int y_exponent, const int *y_fraction,
-            int *z_sign, int *z_exponent, int *z_fraction)
+mp_add_real(const MPNumber *x, int y_sign, const MPNumber *y, MPNumber *z)
 {
     int sign_prod, i, c;
     int exp_diff, med;
-    int x_largest = 0;
+    bool x_largest = false;
     const int *big_fraction, *small_fraction;
-    
-    if (x_sign == 0) {
-        *z_sign = y_sign;
-        *z_exponent = y_exponent;
-        memcpy(z_fraction, y_fraction, sizeof(int) * MP_SIZE);
+    MPNumber x_copy, y_copy;
+
+    /* 0 + y = y */
+    if (mp_is_zero(x)) {
+        mp_set_from_mp(y, z);
+        z->sign = y_sign;
         return;
     }
-    else if (y_sign == 0) {
-        *z_sign = x_sign;
-        *z_exponent = x_exponent;
-        memcpy(z_fraction, x_fraction, sizeof(int) * MP_SIZE);
+    /* x + 0 = x */ 
+    else if (mp_is_zero(y)) {
+        mp_set_from_mp(x, z);
         return;
     }
 
-    sign_prod = y_sign * x_sign;
-    exp_diff = x_exponent - y_exponent;
+    sign_prod = y_sign * x->sign;
+    exp_diff = x->exponent - y->exponent;
     med = abs(exp_diff);
     if (exp_diff < 0) {
-        x_largest = 0;
+        x_largest = false;
     } else if (exp_diff > 0) {
-        x_largest = 1;
+        x_largest = true;
     } else {
         /* EXPONENTS EQUAL SO COMPARE SIGNS, THEN FRACTIONS IF NEC. */
         if (sign_prod < 0) {
             /* Signs are not equal.  find out which mantissa is larger. */
             int j;
             for (j = 0; j < MP_T; j++) {
-                int i = x_fraction[j] - y_fraction[j];
+                int i = x->fraction[j] - y->fraction[j];
                 if (i == 0)
                     continue;
                 if (i < 0)
-                    x_largest = 0;
+                    x_largest = false;
                 else if (i > 0)
-                    x_largest = 1;
+                    x_largest = true;
                 break;
             }
 
             /* Both mantissas equal, so result is zero. */
             if (j >= MP_T) {
-                *z_sign = 0;
-                *z_exponent = 0;
-                memset(z_fraction, 0, sizeof(int) * MP_SIZE);
+                mp_set_from_integer(0, z);
                 return;
             }
         }
     }
 
+    mp_set_from_mp(x, &x_copy);
+    mp_set_from_mp(y, &y_copy);
+    mp_set_from_integer(0, z);
+
     if (x_largest) {
-        *z_sign = x_sign;
-        *z_exponent = x_exponent;
-        big_fraction = x_fraction;
-        small_fraction = y_fraction;
+        z->sign = x_copy.sign;
+        z->exponent = x_copy.exponent;
+        big_fraction = x_copy.fraction;
+        small_fraction = y_copy.fraction;
     } else {
-        *z_sign = y_sign;
-        *z_exponent = y_exponent;
-        big_fraction = y_fraction;
-        small_fraction = x_fraction;
+        z->sign = y_sign;
+        z->exponent = y_copy.exponent;
+        big_fraction = y_copy.fraction;
+        small_fraction = x_copy.fraction;
     }
 
     /* CLEAR GUARD DIGITS TO RIGHT OF X DIGITS */
     for(i = 3; i >= med; i--)
-        z_fraction[MP_T + i] = 0;
+        z->fraction[MP_T + i] = 0;
 
     if (sign_prod >= 0) {
         /* HERE DO ADDITION, EXPONENT(Y) >= EXPONENT(X) */
         for (i = MP_T + 3; i >= MP_T; i--)
-            z_fraction[i] = small_fraction[i - med];
+            z->fraction[i] = small_fraction[i - med];
 
         c = 0;
         for (; i >= med; i--) {
@@ -279,11 +303,11 @@ mp_add_real(int x_sign, int x_exponent, const int *x_fraction,
 
             if (c < MP_BASE) {
                 /* NO CARRY GENERATED HERE */
-                z_fraction[i] = c;
+                z->fraction[i] = c;
                 c = 0;
             } else {
                 /* CARRY GENERATED HERE */
-                z_fraction[i] = c - MP_BASE;
+                z->fraction[i] = c - MP_BASE;
                 c = 1;
             }
         }
@@ -292,92 +316,104 @@ mp_add_real(int x_sign, int x_exponent, const int *x_fraction,
         {
             c = big_fraction[i] + c;
             if (c < MP_BASE) {
-                z_fraction[i] = c;
+                z->fraction[i] = c;
                 i--;
 
                 /* NO CARRY POSSIBLE HERE */
                 for (; i >= 0; i--)
-                    z_fraction[i] = big_fraction[i];
+                    z->fraction[i] = big_fraction[i];
 
-                return;
+                c = 0;
+                break;
             }
 
-            z_fraction[i] = 0;
+            z->fraction[i] = 0;
             c = 1;
         }
 
         /* MUST SHIFT RIGHT HERE AS CARRY OFF END */
         if (c != 0) {
             for (i = MP_T + 3; i > 0; i--)
-                z_fraction[i] = z_fraction[i - 1];
-            z_fraction[0] = 1;
-            (*z_exponent)++;
+                z->fraction[i] = z->fraction[i - 1];
+            z->fraction[0] = 1;
+            z->exponent++;
         }
-
-        return;
     }
-
-    c = 0;
-    for (i = MP_T + med - 1; i >= MP_T; i--) {
-        /* HERE DO SUBTRACTION, ABS(Y) > ABS(X) */
-        z_fraction[i] = c - small_fraction[i - med];
+    else  {
         c = 0;
-
-        /* BORROW GENERATED HERE */
-        if (z_fraction[i] < 0) {
-            c = -1;
-            z_fraction[i] += MP_BASE;
-        }
-    }
-
-    for(; i >= med; i--) {
-        c = big_fraction[i] + c - small_fraction[i - med];
-        if (c >= 0) {
-            /* NO BORROW GENERATED HERE */
-            z_fraction[i] = c;
+        for (i = MP_T + med - 1; i >= MP_T; i--) {
+            /* HERE DO SUBTRACTION, ABS(Y) > ABS(X) */
+            z->fraction[i] = c - small_fraction[i - med];
             c = 0;
-        } else {
+
             /* BORROW GENERATED HERE */
-            z_fraction[i] = c + MP_BASE;
+            if (z->fraction[i] < 0) {
+                c = -1;
+                z->fraction[i] += MP_BASE;
+            }
+        }
+
+        for(; i >= med; i--) {
+            c = big_fraction[i] + c - small_fraction[i - med];
+            if (c >= 0) {
+                /* NO BORROW GENERATED HERE */
+                z->fraction[i] = c;
+                c = 0;
+            } else {
+                /* BORROW GENERATED HERE */
+                z->fraction[i] = c + MP_BASE;
+                c = -1;
+            }
+        }
+
+        for (; i >= 0; i--) {
+            c = big_fraction[i] + c;
+
+            if (c >= 0) {
+                z->fraction[i] = c;
+                i--;
+
+                /* NO CARRY POSSIBLE HERE */
+                for (; i >= 0; i--)
+                    z->fraction[i] = big_fraction[i];
+
+                break;
+            }
+
+            z->fraction[i] = c + MP_BASE;
             c = -1;
         }
     }
 
-    for (; i >= 0; i--) {
-        c = big_fraction[i] + c;
+    mp_normalize(z);
+}
 
-        if (c >= 0) {
-            z_fraction[i] = c;
-            i--;
 
-            /* NO CARRY POSSIBLE HERE */
-            for (; i >= 0; i--)
-                z_fraction[i] = big_fraction[i];
+static void
+mp_add_with_sign(const MPNumber *x, int y_sign, const MPNumber *y, MPNumber *z)
+{
+    if (mp_is_complex(x) || mp_is_complex(y)) {
+        MPNumber real_x, real_y, im_x, im_y, real_z, im_z;
 
-            return;
-        }
+        mp_real_component(x, &real_x);
+        mp_imaginary_component(x, &im_x);
+        mp_real_component(y, &real_y);
+        mp_imaginary_component(y, &im_y);
 
-        z_fraction[i] = c + MP_BASE;
-        c = -1;
+        mp_add_real(&real_x, y_sign * y->sign, &real_y, &real_z);
+        mp_add_real(&im_x, y_sign * y->im_sign, &im_y, &im_z);
+
+        mp_set_from_complex(&real_z, &im_z, z);
     }
+    else
+        mp_add_real(x, y_sign * y->sign, y, z);
 }
 
 
 void
 mp_add(const MPNumber *x, const MPNumber *y, MPNumber *z)
 {
-    /* 0 + y = y
-     * x + 0 = x */
-    if (mp_is_zero(x))
-        mp_set_from_mp(y, z);
-    else if (mp_is_zero(y))
-        mp_set_from_mp(x, z);
-    else {
-        mp_add_real(x->sign, x->exponent, x->fraction,
-                    y->sign, y->exponent, y->fraction,
-                    &z->sign, &z->exponent, z->fraction);
-        mp_normalize(z);
-    }
+    mp_add_with_sign(x, 1, y, z);
 }
 
 
@@ -402,23 +438,35 @@ mp_add_fraction(const MPNumber *x, int64_t i, int64_t j, MPNumber *y)
 void
 mp_subtract(const MPNumber *x, const MPNumber *y, MPNumber *z)
 {
-    /* 0 - y = -y
-     * x - 0 = x */
-    if (mp_is_zero(x)) {
-        mp_set_from_mp(y, z);
-        mp_invert_sign(z, z);
-    }
-    else if (mp_is_zero(y))
-        mp_set_from_mp(x, z);
-    else {
-        mp_add_real(x->sign, x->exponent, x->fraction,
-                    -y->sign, y->exponent, y->fraction,
-                    &z->sign, &z->exponent, z->fraction);
-        mp_normalize(z);
-    }
+    mp_add_with_sign(x, -1, y, z);
 }
 
 
+void
+mp_sgn(const MPNumber *x, MPNumber *z)
+{
+    if (mp_is_zero(x))
+        mp_set_from_integer(0, z);
+    else if (mp_is_negative(x))
+        mp_set_from_integer(-1, z);
+    else
+        mp_set_from_integer(1, z);  
+}
+
+void
+mp_integer_component(const MPNumber *x, MPNumber *z)
+{
+    int i;
+
+    /* Clear fraction */
+    mp_set_from_mp(x, z);
+    for (i = z->exponent; i < MP_SIZE; i++)
+        z->fraction[i] = 0;
+    z->im_sign = 0;
+    z->im_exponent = 0;
+    memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);
+}
+
 void
 mp_fractional_component(const MPNumber *x, MPNumber *z)
 {
@@ -450,13 +498,27 @@ mp_fractional_component(const MPNumber *x, MPNumber *z)
     }
     if (z->fraction[0] == 0)
         z->sign = 0;
+
+    z->im_sign = 0;
+    z->im_exponent = 0;
+    memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);
 }
 
 
 void
-mp_integer_component(const MPNumber *x, MPNumber *z)
+mp_fractional_part(const MPNumber *x, MPNumber *z)
+{
+    MPNumber f;
+    mp_floor(x, &f);
+    mp_subtract(x, &f, z);
+}
+
+
+void
+mp_floor(const MPNumber *x, MPNumber *z)
 {
     int i;
+    bool have_fraction = false, is_negative;
 
     /* Integer component of zero = 0 */
     if (mp_is_zero(x)) {
@@ -470,13 +532,58 @@ mp_integer_component(const MPNumber *x, MPNumber *z)
         return;
     }
 
+    is_negative = mp_is_negative(x);
+
     /* Clear fraction */
     mp_set_from_mp(x, z);
-    for (i = z->exponent; i < MP_SIZE; i++)
+    for (i = z->exponent; i < MP_SIZE; i++) {
+        if (z->fraction[i])
+            have_fraction = true;
         z->fraction[i] = 0;
+    }
+    z->im_sign = 0;
+    z->im_exponent = 0;
+    memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);
+
+    if (have_fraction && is_negative)
+       mp_add_integer(z, -1, z);
 }
 
 
+void
+mp_ceiling(const MPNumber *x, MPNumber *z)
+{
+    MPNumber f;
+
+    mp_floor(x, z);
+    mp_fractional_component(x, &f);
+    if (mp_is_zero(&f))
+        return;
+    mp_add_integer(z, 1, z);
+}
+
+
+void
+mp_round(const MPNumber *x, MPNumber *z)
+{
+    MPNumber f, one;
+    bool do_floor;
+
+    do_floor = !mp_is_negative(x);
+
+    mp_fractional_component(x, &f);
+    mp_multiply_integer(&f, 2, &f);
+    mp_abs(&f, &f);
+    mp_set_from_integer(1, &one);
+    if (mp_is_greater_equal(&f, &one))
+        do_floor = !do_floor;
+
+    if (do_floor)
+        mp_floor(x, z);
+    else
+        mp_ceiling(x, z);
+}
+
 int
 mp_compare_mp_to_mp(const MPNumber *x, const MPNumber *y)
 {
@@ -549,12 +656,11 @@ mp_divide(const MPNumber *x, const MPNumber *y, MPNumber *z)
 }
 
 
-void
-mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
+static void
+mp_divide_integer_real(const MPNumber *x, int64_t y, MPNumber *z)
 {
-    int i__1;
-    int c, i, k, b2, c2, i2, j1, j2, r1;
-    int j11, kh, iq, ir, iqj;
+    int c, i, k, b2, c2, j1, j2;
+    MPNumber x_copy;
 
     /* x/0 */
     if (y == 0) {
@@ -572,34 +678,26 @@ mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
 
     /* Division by -1 or 1 just changes sign */
     if (y == 1 || y == -1) {
-        mp_set_from_mp(x, z);
         if (y < 0)
-            mp_invert_sign(z, z);
+            mp_invert_sign(x, z);
+        else
+            mp_set_from_mp(x, z);
         return;
     }
 
-    /* If dividing by base then can optimise */
-    if (y % MP_BASE == 0) {
-        mp_set_from_mp(x, z);
-        if (y < 0) {
-            z->sign = -z->sign;
-            z->exponent -= -y / MP_BASE;
-        }
-        else
-            z->exponent -= y / MP_BASE;
-        return;
-    }
+    /* Copy x as z may also refer to x */
+    mp_set_from_mp(x, &x_copy);
+    mp_set_from_integer(0, z);
 
     if (y < 0) {
         y = -y;
-        z->sign = -x->sign;
+        z->sign = -x_copy.sign;
     }
     else
-        z->sign = x->sign;
-    z->exponent = x->exponent;
+        z->sign = x_copy.sign;
+    z->exponent = x_copy.exponent;
 
     c = 0;
-    i2 = MP_T + 4;
     i = 0;
 
     /*  IF y*B NOT REPRESENTABLE AS AN INTEGER HAVE TO SIMULATE
@@ -609,16 +707,18 @@ mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
     /* Computing MAX */
     b2 = max(MP_BASE << 3, 32767 / MP_BASE);
     if (y < b2) {
+        int kh, r1;
+
         /* LOOK FOR FIRST NONZERO DIGIT IN QUOTIENT */
         do {
             c = MP_BASE * c;
             if (i < MP_T)
-                c += x->fraction[i];
+                c += x_copy.fraction[i];
             i++;
             r1 = c / y;
             if (r1 < 0)
                 goto L210;
-        } while(r1 == 0);
+        } while (r1 == 0);
 
         /* ADJUST EXPONENT AND GET T+4 DIGITS IN QUOTIENT */
         z->exponent += 1 - i;
@@ -628,7 +728,7 @@ mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
         if (i < MP_T) {
             kh = MP_T + 1 - i;
             for (k = 1; k < kh; k++) {
-                c += x->fraction[i];
+                c += x_copy.fraction[i];
                 z->fraction[k] = c / y;
                 c = MP_BASE * (c - y * z->fraction[k]);
                 i++;
@@ -637,42 +737,39 @@ mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
                 goto L210;
         }
 
-        for (k = kh; k < i2; k++) {
+        for (k = kh; k < MP_T + 4; k++) {
             z->fraction[k] = c / y;
             c = MP_BASE * (c - y * z->fraction[k]);
         }
         if (c < 0)
             goto L210;
 
-        /* NORMALIZE AND ROUND RESULT */
         mp_normalize(z);
-
         return;
     }
 
     /* HERE NEED SIMULATED DOUBLE-PRECISION DIVISION */
     j1 = y / MP_BASE;
+    j2 = y - j1 * MP_BASE;
 
     /* LOOK FOR FIRST NONZERO DIGIT */
     c2 = 0;
-    j2 = y - j1 * MP_BASE;
     do {
         c = MP_BASE * c + c2;
-        i__1 = c - j1;
-        c2 = i < MP_T ? x->fraction[i] : 0;
+        c2 = i < MP_T ? x_copy.fraction[i] : 0;
         i++;
-    } while (i__1 < 0 || (i__1 == 0 && c2 < j2));
+    } while (c < j1 || (c == j1 && c2 < j2));
 
     /* COMPUTE T+4 QUOTIENT DIGITS */
     z->exponent += 1 - i;
     i--;
-    k = 1;
 
     /* MAIN LOOP FOR LARGE ABS(y) CASE */
-    j11 = j1 + 1;
-    while(1) {
+    for (k = 1; k <= MP_T + 4; k++) {
+        int ir, iq, iqj;
+
         /* GET APPROXIMATE QUOTIENT FIRST */
-        ir = c / j11;
+        ir = c / (j1 + 1);
 
         /* NOW REDUCE SO OVERFLOW DOES NOT OCCUR */
         iq = c - ir * j1;
@@ -690,7 +787,7 @@ mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
         }
 
         if (i < MP_T)
-            iq += x->fraction[i];
+            iq += x_copy.fraction[i];
         i++;
         iqj = iq / y;
 
@@ -700,21 +797,10 @@ mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
 
         if (c < 0)
             goto L210;
-
-        ++k;
-        if (k > i2) {
-            mp_normalize(z);
-            
-            if (mp_is_complex(x)) {
-                MPNumber im_z;
-                mp_imaginary_component(x, &im_z);
-                mp_divide_integer(&im_z, y, &im_z);
-                mp_set_from_complex(z, &im_z, z);
-            }
-            return;
-        }
     }
 
+    mp_normalize(z);
+
 L210:
     /* CARRY NEGATIVE SO OVERFLOW MUST HAVE OCCURRED */
     mperr("*** INTEGER OVERFLOW IN MP_DIVIDE_INTEGER, B TOO LARGE ***");
@@ -722,6 +808,23 @@ L210:
 }
 
 
+void
+mp_divide_integer(const MPNumber *x, int64_t y, MPNumber *z)
+{
+    if (mp_is_complex(x)) {
+        MPNumber re_z, im_z;
+
+        mp_real_component(x, &re_z);
+        mp_imaginary_component(x, &im_z);
+        mp_divide_integer_real(&re_z, y, &re_z);
+        mp_divide_integer_real(&im_z, y, &im_z);
+        mp_set_from_complex(&re_z, &im_z, z);
+    }
+    else
+        mp_divide_integer_real(x, y, z);
+}
+
+
 bool
 mp_is_integer(const MPNumber *x)
 {
@@ -733,12 +836,12 @@ mp_is_integer(const MPNumber *x)
     /* This fix is required for 1/3 repiprocal not being detected as an integer */
     /* Multiplication and division by 10000 is used to get around a
      * limitation to the "fix" for Sun bugtraq bug #4006391 in the
-     * mp_integer_component() routine in mp.c, when the exponent is less than 1.
+     * mp_floor() routine in mp.c, when the exponent is less than 1.
      */
     mp_set_from_integer(10000, &t3);
     mp_multiply(x, &t3, &t1);
     mp_divide(&t1, &t3, &t1);
-    mp_integer_component(&t1, &t2);
+    mp_floor(&t1, &t2);
     return mp_is_equal(&t1, &t2);
 
     /* Correct way to check for integer */
@@ -785,7 +888,7 @@ mp_is_natural(const MPNumber *x)
 bool
 mp_is_complex(const MPNumber *x)
 {
-    return false;//x->im_sign != 0;
+    return x->im_sign != 0;
 }
 
 
@@ -948,7 +1051,8 @@ mp_epowy_real(const MPNumber *x, MPNumber *z)
         return;
 
     rz = mp_cast_to_float(z);
-    if ((r__1 = rz - exp(rx), fabs(r__1)) < rz * 0.01f)
+    r__1 = rz - exp(rx);
+    if (fabs(r__1) < rz * 0.01f)
         return;
 
     /*  THE FOLLOWING MESSAGE MAY INDICATE THAT
@@ -970,7 +1074,7 @@ mp_epowy(const MPNumber *x, MPNumber *z)
 
     if (mp_is_complex(x)) {
         MPNumber x_real, r, theta;
-        
+
         mp_real_component(x, &x_real);
         mp_imaginary_component(x, &theta);
 
@@ -1022,7 +1126,7 @@ mp_gcd(int64_t *k, int64_t *l)
 bool
 mp_is_zero(const MPNumber *x)
 {
-    return x->sign == 0; // && x->im_sign == 0;
+    return x->sign == 0 && x->im_sign == 0;
 }
 
 
@@ -1172,13 +1276,13 @@ mp_ln(const MPNumber *x, MPNumber *z)
     }
 
     /* ln(-x) complex */
-    // TEMP: Remove when supporting complex numbers
-    if (mp_is_negative(x)) {
-        /* Translators: Error displayed attempted to take logarithm of negative value */
+    /* FIXME: Make complex numbers optional */
+    /*if (mp_is_negative(x)) {
+        // Translators: Error displayed attempted to take logarithm of negative value
         mperr(_("Logarithm of negative values is undefined"));
         mp_set_from_integer(0, z);
         return;
-    }
+    }*/
 
     if (mp_is_complex(x) || mp_is_negative(x)) {
         MPNumber r, theta, z_real;
@@ -1226,7 +1330,7 @@ mp_is_less_than(const MPNumber *x, const MPNumber *y)
 static void
 mp_multiply_real(const MPNumber *x, const MPNumber *y, MPNumber *z)
 {
-    int c, i, j, xi;
+    int c, i, xi;
     MPNumber r;
 
     /* x*0 = 0*y = 0 */    
@@ -1242,6 +1346,8 @@ mp_multiply_real(const MPNumber *x, const MPNumber *y, MPNumber *z)
     /* PERFORM MULTIPLICATION */
     c = 8;
     for (i = 0; i < MP_T; i++) {
+        int j;
+
         xi = x->fraction[i];
 
         /* FOR SPEED, PUT THE NUMBER WITH MANY ZEROS FIRST */
@@ -1291,15 +1397,15 @@ mp_multiply_real(const MPNumber *x, const MPNumber *y, MPNumber *z)
         }
 
         c = 0;
-        for (j = MP_T + 3; j >= 0; j--) {
-            int ri = r.fraction[j] + c;
+        for (i = MP_T + 3; i >= 0; i--) {
+            int ri = r.fraction[i] + c;
             if (ri < 0) {
                 mperr("*** INTEGER OVERFLOW IN MP_MULTIPLY, B TOO LARGE ***");
                 mp_set_from_integer(0, z);
                 return;
             }
             c = ri / MP_BASE;
-            r.fraction[j] = ri - MP_BASE * c;
+            r.fraction[i] = ri - MP_BASE * c;
         }
 
         if (c != 0) {
@@ -1309,6 +1415,11 @@ mp_multiply_real(const MPNumber *x, const MPNumber *y, MPNumber *z)
         }
     }
 
+    /* Clear complex part */
+    z->im_sign = 0;
+    z->im_exponent = 0;
+    memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);    
+
     /* NORMALIZE AND ROUND RESULT */
     // FIXME: Use stack variable because of mp_normalize brokeness
     for (i = 0; i < MP_SIZE; i++)
@@ -1342,7 +1453,7 @@ mp_multiply(const MPNumber *x, const MPNumber *y, MPNumber *z)
         mp_multiply_real(&real_x, &im_y, &t1);
         mp_multiply_real(&im_x, &real_y, &t2);
         mp_add(&t1, &t2, &im_z);
-        
+
         mp_set_from_complex(&real_z, &im_z, z);
     }
     else {
@@ -1351,11 +1462,11 @@ mp_multiply(const MPNumber *x, const MPNumber *y, MPNumber *z)
 }
 
 
-void
-mp_multiply_integer(const MPNumber *x, int64_t y, MPNumber *z)
+static void
+mp_multiply_integer_real(const MPNumber *x, int64_t y, MPNumber *z)
 {
-    int c, i, c1, c2, j1, j2;
-    int t1, t3, t4, ij, ri = 0, is, ix;
+    int c, i;
+    MPNumber x_copy;
 
     /* x*0 = 0*y = 0 */
     if (mp_is_zero(x) || y == 0) {
@@ -1364,38 +1475,29 @@ mp_multiply_integer(const MPNumber *x, int64_t y, MPNumber *z)
     }
 
     /* x*1 = x, x*-1 = -x */
-    // FIXME: Why is this not working?
+    // FIXME: Why is this not working? mp_ext is using this function to do a normalization
     /*if (y == 1 || y == -1) {
-        mp_set_from_mp(x, z);
-        z->sign *= y;
+        if (y < 0)
+            mp_invert_sign(x, z);
+        else
+            mp_set_from_mp(x, z);
         return;
     }*/
 
-    /* If multiplying by base then can optimise */
-    if (y % MP_BASE == 0) {
-        mp_set_from_mp(x, z);
-        if (y < 0) {
-            z->sign = -z->sign;
-            z->exponent += -y / MP_BASE;
-        }
-        else
-            z->exponent += y / MP_BASE;
-        return;
-    }
+    /* Copy x as z may also refer to x */
+    mp_set_from_mp(x, &x_copy);
+    mp_set_from_integer(0, z);
 
     if (y < 0) {
         y = -y;
-        z->sign = -x->sign;
+        z->sign = -x_copy.sign;
     }
     else
-        z->sign = x->sign;
-    z->exponent = x->exponent + 4;
+        z->sign = x_copy.sign;
+    z->exponent = x_copy.exponent + 4;
 
     /* FORM PRODUCT IN ACCUMULATOR */
     c = 0;
-    t1 = MP_T + 1;
-    t3 = MP_T + 3;
-    t4 = MP_T + 4;
 
     /*  IF y*B NOT REPRESENTABLE AS AN INTEGER WE HAVE TO SIMULATE
      *  DOUBLE-PRECISION MULTIPLICATION.
@@ -1403,31 +1505,36 @@ mp_multiply_integer(const MPNumber *x, int64_t y, MPNumber *z)
 
     /* Computing MAX */
     if (y >= max(MP_BASE << 3, 32767 / MP_BASE)) {
+        int64_t j1, j2;
+
         /* HERE J IS TOO LARGE FOR SINGLE-PRECISION MULTIPLICATION */
         j1 = y / MP_BASE;
         j2 = y - j1 * MP_BASE;
 
         /* FORM PRODUCT */
-        for (ij = 1; ij <= t4; ++ij) {
+        for (i = MP_T + 3; i >= 0; i--) {
+            int64_t c1, c2, is, ix, t;
+
             c1 = c / MP_BASE;
             c2 = c - MP_BASE * c1;
-            i = t1 - ij;
             ix = 0;
-            if (i > 0)
-                ix = x->fraction[i - 1];
-            ri = j2 * ix + c2;
-            is = ri / MP_BASE;
+            if (i > 3)
+                ix = x_copy.fraction[i - 4];
+
+            t = j2 * ix + c2;
+            is = t / MP_BASE;
             c = j1 * ix + c1 + is;
-            z->fraction[i + 3] = ri - MP_BASE * is;
+            z->fraction[i] = t - MP_BASE * is;
         }
     }
     else
     {
-        for (ij = 1; ij <= MP_T; ++ij) {
-            i = t1 - ij;
-            ri = y * x->fraction[i - 1] + c;
+        int64_t ri = 0;
+
+        for (i = MP_T + 3; i >= 4; i--) {
+            ri = y * x_copy.fraction[i - 4] + c;
             c = ri / MP_BASE;
-            z->fraction[i + 3] = ri - MP_BASE * c;
+            z->fraction[i] = ri - MP_BASE * c;
         }
 
         /* CHECK FOR INTEGER OVERFLOW */
@@ -1438,45 +1545,53 @@ mp_multiply_integer(const MPNumber *x, int64_t y, MPNumber *z)
         }
 
         /* HAVE TO TREAT FIRST FOUR WORDS OF R SEPARATELY */
-        for (ij = 1; ij <= 4; ++ij) {
-            i = 5 - ij;
-            ri = c;
-            c = ri / MP_BASE;
-            z->fraction[i - 1] = ri - MP_BASE * c;
+        for (i = 3; i >= 0; i--) {
+            int t;
+
+            t = c;
+            c = t / MP_BASE;
+            z->fraction[i] = t - MP_BASE * c;
         }
     }
 
     /* HAVE TO SHIFT RIGHT HERE AS CARRY OFF END */
-    while(1) {
-        /* NORMALIZE AND ROUND OR TRUNCATE RESULT */
-        if (c == 0)
-        {
-            mp_normalize(z);
-            return;
-        }
+    while (c != 0) {
+        int64_t t;
 
-        if (c < 0) {
-            mperr("*** INTEGER OVERFLOW IN mp_multiply_integer, B TOO LARGE ***");
-            mp_set_from_integer(0, z);
-            return;
-        }
-
-        for (ij = 1; ij <= t3; ++ij) {
-            i = t4 - ij;
+        for (i = MP_T + 3; i >= 1; i--)
             z->fraction[i] = z->fraction[i - 1];
-        }
-        ri = c;
-        c = ri / MP_BASE;
-        z->fraction[0] = ri - MP_BASE * c;
+        t = c;
+        c = t / MP_BASE;
+        z->fraction[0] = t - MP_BASE * c;
         z->exponent++;
     }
-    
-    if (mp_is_complex(x)) {
-        MPNumber im_z;
-        mp_imaginary_component(x, &im_z);
-        mp_multiply_integer(&im_z, y, &im_z);
-        mp_set_from_complex(z, &im_z, z);
+
+    if (c < 0) {
+        mperr("*** INTEGER OVERFLOW IN mp_multiply_integer, B TOO LARGE ***");
+        mp_set_from_integer(0, z);
+        return;
     }
+
+    z->im_sign = 0;
+    z->im_exponent = 0;
+    memset(z->im_fraction, 0, sizeof(int) * MP_SIZE);    
+    mp_normalize(z);
+}
+
+
+void
+mp_multiply_integer(const MPNumber *x, int64_t y, MPNumber *z)
+{
+    if (mp_is_complex(x)) {
+        MPNumber re_z, im_z;
+        mp_real_component(x, &re_z);
+        mp_imaginary_component(x, &im_z);
+        mp_multiply_integer_real(&re_z, y, &re_z);
+        mp_multiply_integer_real(&im_z, y, &im_z);
+        mp_set_from_complex(&re_z, &im_z, z);
+    }
+    else
+        mp_multiply_integer_real(x, y, z);
 }
 
 
@@ -1506,7 +1621,7 @@ mp_invert_sign(const MPNumber *x, MPNumber *z)
 {
     mp_set_from_mp(x, z);
     z->sign = -z->sign;
-    //z->im_sign = -z->im_sign;
+    z->im_sign = -z->im_sign;
 }
 
 
@@ -1548,11 +1663,12 @@ mp_pwr(const MPNumber *x, const MPNumber *y, MPNumber *z)
     MPNumber t;
 
     /* (-x)^y imaginary */
-    if (x->sign < 0) {
+    /* FIXME: Make complex numbers optional */
+    /*if (x->sign < 0) {
         mperr(_("The power of negative numbers is only defined for integer exponents"));
         mp_set_from_integer(0, z);
         return;
-    }
+    }*/
 
     /* 0^-y illegal */
     if (mp_is_zero(x) && y->sign < 0) {
@@ -1651,8 +1767,8 @@ mp_reciprocal(const MPNumber *x, MPNumber *z)
 }
 
 
-void
-mp_root(const MPNumber *x, int64_t n, MPNumber *z)
+static void
+mp_root_real(const MPNumber *x, int64_t n, MPNumber *z)
 {
     float approximation;
     int ex, np, it0, t;
@@ -1756,23 +1872,45 @@ mp_root(const MPNumber *x, int64_t n, MPNumber *z)
 }
 
 
+void
+mp_root(const MPNumber *x, int64_t n, MPNumber *z)
+{
+    if (!mp_is_complex(x) && mp_is_negative(x) && n % 2 == 1) {
+        mp_abs(x, z);
+        mp_root_real(z, n, z);
+        mp_invert_sign(z, z);
+    }
+    else if (mp_is_complex(x) || mp_is_negative(x)) {
+        MPNumber r, theta;
+
+        mp_abs(x, &r);
+        mp_arg(x, MP_RADIANS, &theta);
+
+        mp_root_real(&r, n, &r);
+        mp_divide_integer(&theta, n, &theta);
+        mp_set_from_polar(&r, MP_RADIANS, &theta, z);
+    }
+    else
+        mp_root_real(x, n, z);
+}
+
+
 void
 mp_sqrt(const MPNumber *x, MPNumber *z)
 {
-    if (x->sign < 0) {
+    if (mp_is_zero(x))
+       mp_set_from_integer(0, z);
+    /* FIXME: Make complex numbers optional */
+    /*else if (x->sign < 0) {
         mperr(_("Square root is undefined for negative values"));
         mp_set_from_integer(0, z);
-    }
-    else if (mp_is_zero(x))
-        mp_set_from_integer(0, z);
+    }*/
     else {
-        int i;
         MPNumber t;
 
         mp_root(x, -2, &t);
-        i = t.fraction[0]; // ?
         mp_multiply(x, &t, z);
-        mp_ext(i, z->fraction[0], z);
+        mp_ext(t.fraction[0], z->fraction[0], z);
     }
 }
 
@@ -1814,7 +1952,7 @@ mp_modulus_divide(const MPNumber *x, const MPNumber *y, MPNumber *z)
     }
 
     mp_divide(x, y, &t1);
-    mp_integer_component(&t1, &t1);
+    mp_floor(&t1, &t1);
     mp_multiply(&t1, y, &t2);
     mp_subtract(x, &t2, z);
 
@@ -1865,6 +2003,12 @@ mp_xpowy_integer(const MPNumber *x, int64_t n, MPNumber *z)
         mp_set_from_integer(0, z);
         return;
     }
+  
+    /* x^1 = x */
+    if (n == 1) {
+        mp_set_from_mp(x, z);
+        return;
+    }
 
     if (n < 0) {
         mp_reciprocal(x, &t);
