@@ -128,14 +128,14 @@ test_conversions()
 }
 
 
-int
+static int
 variable_is_defined(const char *name, void *data)
 {
     return strcmp (name, "x") == 0 || strcmp (name, "y") == 0;
 }
 
 
-int
+static int
 get_variable(const char *name, MPNumber *z, void *data)
 {
     if (strcmp (name, "x") == 0) {
@@ -150,12 +150,12 @@ get_variable(const char *name, MPNumber *z, void *data)
 }
 
 
-void
+static void
 set_variable(const char *name, const MPNumber *x, void *data)
 {
 }
 
-void
+static void
 test_equations()
 {
     memset(&options, 0, sizeof(options));
@@ -400,6 +400,9 @@ test_equations()
     test("8 mod 7", "1", 0);
     test("−1 mod 7", "6", 0);
 
+    test("sgn 0", "0", 0);  
+    test("sgn 3", "1", 0);
+    test("sgn −3", "−1", 0);
     test("⌊3⌋", "3", 0);
     test("⌈3⌉", "3", 0);
     test("[3]", "3", 0);
@@ -534,6 +537,8 @@ test_equations()
     test("e^iπ", "−1", 0);
     test("log (−10) − (1 + πi÷ln(10))", "0", 0);
     test("ln (−e) − (1 + πi)", "0", 0);
+    test("sin(iπ÷4) − i×sinh(π÷4)", "0", 0);
+    test("cos(iπ÷4) − cosh(π÷4)", "0", 0);
 
     /* Boolean */
     test("0 and 0", "0", 0);
